# Defines functions: tp, tp_loc, tp_glob, tp_it, tp_itglob

################################################################################

tp <- function(data, indexes = 1:dim(data)[1], k, dnoise, sigma, n, loc = FALSE,
               verbose = FALSE) {
  # 'indexes' give the indexes for which local dimension estimation should
  # be performed.
  # 'k' is the number of neighbors used for each local dimension estimation.
  # 'dnoise' is a vector valued function giving the transition density.
  # 'sigma' is the estimated standard deviation for the noise.
  # 'n' is the dimension of the noise (at least dim(data)[2])

  N <- length(indexes)
  nbh.dist <- yaIkNN(data, indexes, k)[ , (k + 1):(2*k)]
  mi <- rep(NA, N) # This vector will hold local dimension estimates

  for (i in 1:N) {
    Rs <- nbh.dist[i, ]
    mi[i] <- tpR(Rs, dnoise, sigma, n)
  }
  if (verbose) {
    message("Local dimension estimates:\n")
    print(mi)
    message("Mean:", mean(mi), "\n")
  }
  
  if (loc) return(cbind(de = mi, likelihood = NA))
  return(c(de = N/sum(1/mi), likelihood = NA))
  
}

tp_loc <- function(data, dnoise, sigma, n) {
  center <- apply(data, 2, mean)
  cent.data <- t(t(data) - center)
  return(c(de = tpR(sort(lens(cent.data)), dnoise, sigma, n), likelihood = NA))
}

################################################################################

tp_glob <- function(data, k, dnoise, sigma, n) {
  # 'k' is the number of distances used for dimension estimation.
  # 'dnoise' is a vector valued function giving the transition density.
  # 'sigma' is the estimated standard deviation for the noise.
  # 'n' is the dimension of the noise (at least dim(data)[2])

  N <- dim(data)[1]
  dist <- yaIkNN(data, 1:N, 5)[, 6:10]
  dist <- dist[!duplicated(dist[1:(5*N)])]
  Rs <- dist[rank(dist, ties.method = 'random') <= k]
  return(c(de = tpR(Rs, dnoise, sqrt(2)*sigma, n), likelihood = NA)) # Since distances between points
                                            # are used, noise is added at both
                                            # ends, i.e. variance is doubled. 

}

################################################################################

tp_it <- function(data, indexes = 1:dim(data)[1], k, dnoise, sigma, n, init = 5,
                  loc = FALSE, verbose = FALSE) {

  N <- length(indexes)
  nbh.dist <- yaIkNN(data, indexes, k)[ , (k + 1):(2*k)]
  mi <- rep(NA, N) # This vector will hold local dimension estimates
  for (i in 1:N) {
    Rs <- nbh.dist[i, ]
    mi[i] <- tpitR(Rs, dnoise, sigma, n, init)
  }
  if (verbose) {
    cat("Local dimension estimates:\n")
    print(mi)
    cat("Mean:", mean(mi), "\n")
  }
  
  if (loc) return(mi)
  return(c(de = N/sum(1/mi), likelihood = NA))
  
}

################################################################################

tp_itglob <- function(data, k, dnoise, sigma, n, init = 5) {
  # 'k' is the number of distances used for dimension estimation.
  # 'dnoise' is a vector valued function giving the transition density.
  # 'sigma' is the estimated standard deviation for the noise.
  # 'n' is the dimension of the noise (at least dim(data)[2])

  N <- dim(data)[1]
  dist <- yaIkNN(data, 1:N, 5)[, 6:10]
  dist <- dist[!duplicated(dist[1:(5*N)])]
  Rs <- dist[rank(dist, ties.method = 'random') <= k]
  return(c(de = tpitR(Rs, dnoise, sqrt(2)*sigma, n, init), likelihood = NA))

}

################################################################################

tpR <- function(Rs, dnoise, sigma, n) {

  if (!typeof(dnoise) == 'closure') dnoise <- match.fun(dnoise)
  k <- length(Rs)
  Rk <- max(Rs)
  if (sigma == 0) return((k - 1)/(sum(log(Rk/Rs))))
  
  Rpr <- Rk + 100*sigma

  numerator <- rep(NA, k - 1)
  denominator <- rep(NA, k - 1)

  for (j in 1:(k-1)) {

    Rj <- Rs[j]
    tc <- tryCatch({
      numInt <- integrate(function(x) {
                            dnoise(x, Rj, sigma, n) * log(Rk/x)
                          }, 0, Rpr, rel.tol = 1e-2)
    }, error = function(ex) {
       print(ex)
       return(NA)
    }, finally = {})
    if (is.na(tc)[1]) return(NA)
    numerator[j] <- numInt$value

    tc <- tryCatch({
      denomInt <- integrate(function(x) {
                              dnoise(x, Rj, sigma, n)
                            }, 0, Rpr, rel.tol = 1e-2)
    }, error = function(ex) {
      print(ex)
      return(NA)
    }, finally = {})
    if (is.na(tc)[1]) return(NA)
    denominator[j] <- denomInt$value
  }

  return((k-1)/sum(numerator/denominator)) # k-2 instead of k-1 gives
                                           # unbiased estimator if no noise
                                           # or boundary

}

################################################################################

tpitR <- function(Rs, dnoise, sigma, n, init = 5, verbose = FALSE) {

  if (!typeof(dnoise) == 'closure') dnoise <- match.fun(dnoise)
  m <- init
  for (i in 1:5) {
    m <- tpitR_inner(Rs, dnoise, sigma, n, m)
    if (verbose) cat('Iteration', i, ':', m, '\n')
  }
  return(m)

}

tpitR_inner <- function(Rs, dnoise, sigma, n, m) {

  k <- length(Rs)                        
  Rk <- max(Rs)
  if (sigma == 0) return((k - 1)/(sum(log(Rk/Rs))))
  Rpr <- Rk + 100*sigma

  numerator <- rep(NA, k - 1)
  denominator <- rep(NA, k - 1)

  for (j in 1:(k-1)) {
    Rj <- Rs[j]
    numInt <- integrate(function(x) {
                          m <- max(m, 1)
                          x^(m-1)*dnoise(x, Rj, sigma, n) * log(Rk/x)
                        }, 0, Rpr, rel.tol = 1e-2)
    numerator[j] <- numInt$value

    denomInt <- integrate(function(x) {
                            m <- max(m, 1)
                            x^(m-1)*dnoise(x, Rj, sigma, n)
                          }, 0, Rpr, rel.tol = 1e-2)
    denominator[j] <- denomInt$value
  }

  return((k-1)/sum(numerator/denominator)) # k-2 instead of k-1 gives
                                           # unbiased estimator if no noise
                                           # or boundary

}

################################################################################

yaIkNN <- function(data, indexes, k) {
  # ann returns indexes and _squared_ Euclidean distances
  kNN.i.d <- ann(data, data[indexes, , drop = FALSE], k + 1, 
                    verbose = FALSE)$knnIndexDist[ , -c(1, k + 2), drop = FALSE]
  kNN.d <- sqrt(kNN.i.d[ , (k + 1):(2*k)])
  kNN.i.d[ , (k + 1):(2*k)] <- kNN.d
  return(kNN.i.d)
}

################################################################################

